/* 
 * Builds a dictionary from tile ids and textures.
 * Each type can have multiple textures.
 */
#ifndef TILE_DICTIONARY_H
#define TILE_DICTIONAR_H

#include "vul_types.h"

#include "tile_dictionary.h"

int ld_load_image_to_texture( sl_texture *tex_out, const char *path_to_image )
{
	unsigned char *data;
	int width, height, component_count;
	
	data = stbi_load( path_to_image, &width, &height, &component_count, 4 );

	if( data != NULL )
	{
		sl_texture_create( tex_out, data, width, height, GL_RGBA, GL_RGBA8 );
		return 0;
	} else {
		return 1;
	}
}

ui32_t ld_ui32t_modulo_hash( const ui8_t *data, ui32_t len )
{
	assert( len == sizeof( ui32_t ) );

	return *( ( ui32_t* )data );
}
int ld_list_ui32t_comparator( void *a, void *b )
{
	ui32_t *ia, *ib;
	
	ia = ( ui32_t* )a;
	ib = ( ui32_t* )b;

	return *ia - *ib;
}
int ld_map_ui32t_comparator( void *a, void *b )
{
	vul_hash_map_element_t *ea, *eb;
	ui32_t *ia, *ib;
	
	ea = ( vul_hash_map_element_t* )a;
	eb = ( vul_hash_map_element_t* )b;

	ia = ( ui32_t* )ea->key;
	ib = ( ui32_t* )eb->key;

	return *ia - *ib;
}

vul_list_element_t *ld_tile_dictionary_add_key( ui32_t key )
{
	vul_list_element_t *e;

	assert( ld_tile_dictionary_instance );

	e = vul_list_insert( ld_tile_dictionary_instance->keys, &key, sizeof( ui32_t ), ld_list_ui32t_comparator );
	if( e->prev == NULL ) { // It's the new first element, store the new head!
		ld_tile_dictionary_instance->keys = e;
	}

	return e;
}

void ld_tile_dictionary_add_entry( vul_list_element_t *key_entry,
								   ui32_t var_count, ui32_t frame_count, ui32_t frame_freq,
								   ui32_t max_swaps, char *base_name )
{
	ui32_t i, j;
	vul_hash_map_element_t *e, element;
	ld_tile_dictionary_entry *entry;
	char path[ 256 ];
	sl_box uvs;
	sl_vec delta;
	sl_animation_sprite_state *as;
	sl_texture *t;
	
	assert( ld_tile_dictionary_instance );

	entry = ( ld_tile_dictionary_entry* )malloc( sizeof( ld_tile_dictionary_entry ) );
	entry->variation_count = var_count;
	entry->max_swaps = max_swaps;
	entry->texture_ids = ( ui32_t* )malloc( sizeof( ui32_t ) * var_count );
	entry->animation_states = ( vul_vector_t** )malloc( sizeof( vul_vector_t* ) * var_count );
	entry->inverse_frequency = frame_freq;
	
	// For every variation
	for( i = 0; i < var_count; ++i ) {
		// Texture
		snprintf( path, 255, "../res/tiles/%s%d.png", base_name, ( int )i ); // @TODO: Add option for blurry textures, add a b after the number
		t = sl_renderer_allocate_texture( );
		ld_load_image_to_texture( t, path );
		entry->texture_ids[ i ] = t->texture_id;
		// Animation
		if( frame_count > 1 ) {
			entry->animation_states[ i ] = vul_vector_create( sizeof( sl_animation_sprite_state ), frame_count );
			sl_vset( &delta, 1.f / sqrtf( frame_count ), 1.f / sqrtf( frame_count ) );
			sl_bset_scalar( &uvs, 0.f, 0.f, delta.x, delta.y );
			for( j = 0; j < frame_count; ++j ) {
				as = ( sl_animation_sprite_state* )vul_vector_add_empty( entry->animation_states[ i ] );
				as->texture_id = entry->texture_ids[ i ];
				sl_bset( &as->uvs, &uvs );
				sl_bset_scalar( &uvs, uvs.min_p.x + delta.x, uvs.min_p.y,
									  uvs.max_p.x + delta.x, uvs.max_p.y );
				if( uvs.min_p.x >= 0.99999f ) {
					sl_bset_scalar( &uvs, 0.f,	   uvs.min_p.y + delta.y,
										  delta.x, uvs.max_p.y + delta.y );
				}
			}
		} else {
			entry->animation_states[ i ] = NULL;
		}
	}

	// Insert it into the map
	e = vul_map_get( ld_tile_dictionary_instance->mapping, ( ui8_t* )&key_entry->data, key_entry->data_size );
	if( e != NULL ) {
		// Delete the old one and insert the new one
		free( ( ( ld_tile_dictionary_entry* )e->data )->texture_ids );
		for( i = 0; i < var_count; ++i ) {
			vul_vector_destroy( ( ( ld_tile_dictionary_entry* )e->data )->animation_states[ i ] );
		}
		free( ( ( ld_tile_dictionary_entry* )e->data )->animation_states );
		memcpy( e->data, &entry, sizeof( entry ) );
	} else {
		// Create the element
		element.data = entry;
		element.data_size = sizeof( ld_tile_dictionary_entry );
		element.key = ( ui8_t* )key_entry->data;
		element.key_size = key_entry->data_size;

		// Insert it
		e = vul_map_insert( ld_tile_dictionary_instance->mapping, &element );
	}
}

void ld_tile_dictionary_finalize( )
{
	vul_list_element_t *e;
	ld_tile_dictionary_entry *entry;
	ui32_t i;

	for( i = 0; i < ld_tile_dictionary_instance->mapping->bucket_count; ++i ) {
		e = ld_tile_dictionary_instance->mapping->buckets[ i ];
		while( e != NULL ) {
			entry = ( ld_tile_dictionary_entry* )e->data;
			free( entry->texture_ids );
			for( i = 0; i < entry->variation_count; ++i ) {
				vul_vector_destroy( ( ( ld_tile_dictionary_entry* )e->data )->animation_states[ i ] );
			}
			free( entry->animation_states );
			free( entry );
		}
	}

	vul_list_destroy( ld_tile_dictionary_instance->keys );
	vul_map_destroy( ld_tile_dictionary_instance->mapping );
}

void ld_tile_read_dictionary( )
{
	FILE *f;
	char line[ 256 ];
	char base_name[ 236 ];
	char *full_file, *c, *cl;
	size_t file_len, read_len;
	ui32_t id_count, var_count, frame_count, frame_freq;
	int swaps;
	vul_list_element_t *le;

	// Create the dictionary
	ld_tile_dictionary_instance = ( ld_tile_dictionary* )malloc( sizeof( ld_tile_dictionary ) );
	ld_tile_dictionary_instance->keys = NULL;
	ld_tile_dictionary_instance->mapping = vul_map_create( TILE_DICTIONARY_BUCKET_COUNT, ld_ui32t_modulo_hash, ld_map_ui32t_comparator );

	// Open the listing
	f = fopen( "../res/tiles.lst", "r" );
	
	// Find file length
	fseek( f, 0, SEEK_END );
	file_len = ftell( f );
	fseek( f, 0, SEEK_SET );
	
	// Read the entire file
	full_file = ( char* )malloc( file_len + 1 );
	read_len = fread( full_file, 1, file_len, f );
	full_file[ file_len ] = 0;
	if( read_len != file_len ) {
		printf( "File length and read length disagree. %u vs %u.\n", file_len, read_len );
	}
	fclose( f );

	// Scan the file
	c = full_file;
	cl = c;
	while( *c ) {
		if( *c == '\n' ) {
			// Scan the line
			strncpy( line, cl, c - cl );
			line[ c - cl ] = 0;
			if( line[ 0 ] != '#' ) { // Skip comments
				sscanf( line, "%u|%u|%u|%u|%d|%s", &id_count, &var_count, &frame_count, &frame_freq, &swaps, &base_name );
				le = ld_tile_dictionary_add_key( id_count );
				ld_tile_dictionary_add_entry( le, var_count, frame_count, frame_freq,
											  swaps == -1 ? TILE_INFINITELY_SWAPPABLE : 
															swaps == -2 ? TILE_PASSIVLEY_SWAPPABLE :
																		  ( ui32_t )swaps, base_name );
			}
			cl = c + 1;
		}
		++c;
	}
}

void ld_tile_dictionary_get( sl_texture **tex, vul_vector_t **anim_states, sl_box *uvs, f32_t *freq, ui32_t *max_swaps, ui32_t tile_id )
{
	int r;
	ui32_t frame_count;
	const vul_hash_map_element_t *e;
	ld_tile_dictionary_entry *entry;

	e = vul_map_get_const( ld_tile_dictionary_instance->mapping, ( ui8_t* )&tile_id, sizeof( ui32_t ) );
	if( e == NULL ) {
		assert( SL_FALSE ); // Asked for a tile that doesn't exist
		tex = NULL;
		anim_states = NULL;
	}
	assert( e->data_size == sizeof( ld_tile_dictionary_entry ) );

	entry = ( ld_tile_dictionary_entry* )e->data;
	r = ( int )( ( ( float )rand( ) / ( float )RAND_MAX ) * ( float )entry->variation_count );

	*tex = sl_renderer_get_texture_by_id( entry->texture_ids[ r ] );
	*anim_states = entry->animation_states[ r ];
	*max_swaps = entry->max_swaps;

	if( entry->animation_states[ r ] != NULL ) {
		*freq = 1.f / entry->inverse_frequency;
		frame_count = vul_vector_size( entry->animation_states[ r ] );
		sl_bset_scalar( uvs, 0.0f, 0.0f,
						1.f / sqrtf( ( f32_t )frame_count ),
						1.f / sqrtf( ( f32_t )frame_count ) );
	} else {
		sl_bset_scalar( uvs, 0.0f, 0.0f, 1.f, 1.f ); // No animation means full uvs
	}
}

ui32_t ld_get_tile_swaps_by_texture_id( unsigned int tex_id )
{
	ld_tile_dictionary_entry *e;
	vul_list_element_t *le;
	const vul_hash_map_element_t *he;
	ui32_t i;

	// @TODO: Save this mapping directly, this is slow as fuck
	le = ld_tile_dictionary_instance->keys;
	while( le )
	{
		he = vul_map_get_const( ld_tile_dictionary_instance->mapping, ( ui8_t* )le->data, le->data_size );
		assert( he != NULL ); // This should never happen!
		assert( he->data_size == sizeof( ld_tile_dictionary_entry ) ); // Should be a sure thing
		e = ( ld_tile_dictionary_entry* )he->data;
		for( i = 0; i < e->variation_count; ++i ) {
			if( e->texture_ids[ i ] == tex_id ) {
				return e->max_swaps;
			}
		}
		le = le->next;
	}
	
	assert( SL_FALSE ); // We have looked for a non-existant tile!
	return 0;
}

#endif
